"""Dialogue with a single line text input field.
"""
# imports
from kivy.properties import StringProperty
from dialogue import Dialogue
from kivy.lang import Builder
from os.path import join, dirname

# exports
__all__ = ('TextInputDialogue', )

# Load kv
Builder.load_file(join(dirname(__file__), 'dlg_input.kv'))

# Classes
class TextInputDialogue(Dialogue):
    """Dialogue with a single line text input field.

    **keyboard_ctrl** is an object supporting *request_exclusive_keyboard* and
    *release_exclusive_keyboard* for exclusive keyboard access. Usually the *root_widget*.

    Pass the default text as **text**.

    >>> TextInputDialogue(text='Hello world', keyboard_ctrl=root_widget).open()

    In case of touch events, they need to be inhibited to change the focus.

    >>> FocusBehavior.ignored_touch.append(touch)

    """

    # Defocus problem:
    # Buttons defocus when on_press, but on_release is ok.
    # Touch events must be blocked via FocusBehavior

    text = StringProperty('')

    def __init__(self, **kwargs):
        self._keyboard_ctrl = kwargs.pop('keyboard_ctrl', None)
        super(TextInputDialogue, self).__init__(**kwargs)
        if self._keyboard_ctrl is not None:
            self._keyboard_ctrl.request_exclusive_keyboard()

    def on_text_focus(self, instance, value):
        if not value:
            self.cancel()

    def ok(self):
        if self._keyboard_ctrl is not None: # release keyboard
            self._keyboard_ctrl.release_exclusive_keyboard()
        super(TextInputDialogue, self).ok()

    def cancel(self):
        if self._keyboard_ctrl is not None: # release keyboard
            self._keyboard_ctrl.release_exclusive_keyboard()
        super(TextInputDialogue, self).cancel()

## EOF ##
